#include "Triangle.h"
#include "DOFAdmin.h"
#include "Mesh.h"
#include "CoarseningManager.h"
#include "FixVec.h"
#include "ElementDofIterator.h"

namespace AMDiS {

  const int Triangle::vertexOfEdge[3][2] = {{1, 2}, {2, 0}, {0, 1}};
  const int Triangle::sideOfChild[2][3] = {{-1, 2, 0}, {2, -1, 1}};
  const int Triangle::vertexOfParent[2][3] = {{2, 0, -1}, {1, 2, -1}};

  bool Triangle::hasSide(Element* sideElem) const
  {
    FUNCNAME("Triangle::hasSide");
    TEST_EXIT_DBG(sideElem->isLine())("called for sideElem-type != Line\n");
    ERROR_EXIT("not yet\n");
    return false;
  }

  int Triangle::getVertexOfPosition(GeoIndex position, int positionIndex,
				    int vertexIndex) const 
  {
    FUNCNAME("Triangle::getVertexOfPosition");
    switch(position) {
    case VERTEX:
      return positionIndex;
      break;
    case EDGE:
      return vertexOfEdge[positionIndex][vertexIndex];
      break;
    case CENTER:
      return vertexIndex;
      break;
    default:
      ERROR_EXIT("invalid position\n");
      return 0;
    }
  }

  const FixVec<int, WORLD>& Triangle::sortFaceIndices(int face, 
						      FixVec<int,WORLD> *vec) const
  {
    static MatrixOfFixVecs<FixVec<int,WORLD> > *sorted_2d = NULL;

    int no = 0;
    FixVec<int,WORLD> *val = NULL;
    const int *vof = vertexOfEdge[face];

    if (NULL == sorted_2d) {
      sorted_2d = new MatrixOfFixVecs<FixVec<int,WORLD> >(2, 3, 2, NO_INIT);

      (*sorted_2d)[1][0][1] = (*sorted_2d)[1][1][0] =
	(*sorted_2d)[2][0][0] = (*sorted_2d)[2][1][1] = 0;
      (*sorted_2d)[0][0][0] = (*sorted_2d)[0][1][1] =
	(*sorted_2d)[2][0][1] = (*sorted_2d)[2][1][0] = 1;
      (*sorted_2d)[0][0][1] = (*sorted_2d)[0][1][0] =
	(*sorted_2d)[1][0][0] = (*sorted_2d)[1][1][1] = 2;
    }

    if (dof[vof[0]][0] < dof[vof[1]][0])
      no = 0;
    else
      no = 1;

    if (vec) {
      val = vec;
      *val = (*sorted_2d)[face][no];
    } else {
      val = &((*sorted_2d)[face][no]);
    }

    return *(const_cast<const FixVec<int,WORLD>* >(val));
  }


  void Triangle::getVertexDofs(FiniteElemSpace* feSpace, 
			       BoundaryObject bound,
			       DofContainer& dofs, 
			       bool parentVertices) const
  {
    FUNCNAME("Triangle::getVertexDofs()");

    if (parentVertices) {
      switch (bound.ithObj) {
      case 0:
	dofs.push_back(dof[1]);
	break;
      case 1:
	dofs.push_back(dof[0]);
	break;
      case 2:
	dofs.push_back(dof[0]);
	break;
      default:
	ERROR_EXIT("Should never happen!\n");
      }
    }

    BoundaryObject nextBound = bound;

    switch (bound.ithObj) {
    case 0: 
      {
	Element *child1 = child[1];
	if (child1 && child1->getFirstChild()) {
	  nextBound.ithObj = 0;
	  child1->getFirstChild()->getVertexDofs(feSpace, nextBound, dofs);

	  dofs.push_back(child1->getFirstChild()->getDOF(2));

	  nextBound.ithObj = 1;
	  child1->getSecondChild()->getVertexDofs(feSpace, nextBound, dofs);
	}
      }
      break;
    case 1:
      {
	Element *child0 = child[0];
	if (child0 && child0->getFirstChild()) {
	  nextBound.ithObj = 0;
	  child0->getFirstChild()->getVertexDofs(feSpace, nextBound, dofs);

	  dofs.push_back(child[0]->getFirstChild()->getDOF(2));

	  nextBound.ithObj = 1;
	  child0->getSecondChild()->getVertexDofs(feSpace, nextBound, dofs);
	}
      }
      break;
    case 2:      
      if (child[0]) {
	nextBound.ithObj = 0;
	child[0]->getVertexDofs(feSpace, nextBound, dofs);

	dofs.push_back(child[0]->getDOF(2));

	nextBound.ithObj = 1;
	child[1]->getVertexDofs(feSpace, nextBound, dofs);
      }
      break;      
    default:
      ERROR_EXIT("Should never happen!\n");
    }

    if (parentVertices) {
      switch (bound.ithObj) {
      case 0:
	dofs.push_back(dof[2]);
	break;
      case 1:
	dofs.push_back(dof[2]);
	break;
      case 2:
	dofs.push_back(dof[1]);
	break;
    default:
      ERROR_EXIT("Should never happen!\n");
      }
    }
  }


  void Triangle::getNonVertexDofs(FiniteElemSpace* feSpace, 
				  BoundaryObject bound,
				  DofContainer& dofs) const
  {
    FUNCNAME("Triange::getNonVertexDofs()");

    bool addThisEdge = false;
    BoundaryObject nextBound = bound;

    switch (bound.ithObj) {
    case 0:
      if (child[1]) {
	nextBound.ithObj = 2;
	child[1]->getNonVertexDofs(feSpace, nextBound, dofs);
      } else {
	addThisEdge = true;
      }

      break;
    case 1:
      if (child[0]) {
	nextBound.ithObj = 2;
	child[0]->getNonVertexDofs(feSpace, nextBound, dofs);
      } else {
	addThisEdge = true;
      }

      break;
    case 2:
      if (child[0]) {
	nextBound.ithObj = 0;
	child[0]->getNonVertexDofs(feSpace, nextBound, dofs);
	nextBound.ithObj = 1;
	child[1]->getNonVertexDofs(feSpace, nextBound, dofs);
      } else {
	addThisEdge = true;
      } 

      break;
    default:
      ERROR_EXIT("Should never happen!\n");
    }

    if (addThisEdge) {
      ElementDofIterator elDofIter(feSpace, true);
      elDofIter.reset(this);
      do {
	if (elDofIter.getCurrentPos() == 1 && 
	    elDofIter.getCurrentElementPos() == bound.ithObj)
	  dofs.push_back(elDofIter.getDofPtr());	
      } while(elDofIter.next());      
    }
  }


}
